/** \file
 \brief Bin to DFU converter
 \author Tomasz Ostrowski, http://tomeko.net
 \date May 2013

 Basic .bin to .dfu command-line converter. Tested with STM32F303 board.
*/
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <stdint.h>
#include <getopt.h>

/** Flag set by `--verbose'. */
static int verbose_flag;

char fin[255]; 			///< input filename
char fout[255]; 		///< output filename
unsigned int starting_address;

enum { VID = 0x0483 };
enum { PID = 0xdf11 };
enum { VER = 0x2200 };

#define TARGET_PREFIX_SIGNATURE_SIZE	6
#define TARGET_PREFIX_SIGNATURE		"Target"
typedef struct
{
	char	szSignature[TARGET_PREFIX_SIGNATURE_SIZE];
	unsigned char bAlternateSetting;
	int 	bTargetNamed;
	char	szTargetName[255];
	unsigned int	dwTargetSize;
	unsigned int	dwNbElements;
} __attribute__ ((packed)) TARGETPREFIX;

FILE* fileOut;
unsigned int crc = 0xFFFFFFFF;
extern "C" uint32_t crc32(uint32_t crc, const void *buf, uint32_t size);
void write(const void* buf, int size)
{
    int ret = fwrite(buf, size, 1, fileOut);
    crc = crc32(crc, buf, size);

    if (ret != 1)
    {
        puts("Error writing output file\n");
        fclose(fileOut);
        exit (1);
    }
}

void write(unsigned char data)
{
    write(&data, sizeof(data));
}

void write(unsigned short data)
{
    write(&data, sizeof(data));
}

void write(unsigned int data)
{
    write(&data, sizeof(data));
}


/** \brief Print some help */
void help(char *exename)
{
    puts("Crappy bin to DFU file converter for STM32F3\n");
    printf("Usage: %s --i <input_filename> --a <hex_starting_address> --o <output_filename>\n", exename);
    printf(" i.e. %s --i STM32_test.bin --a 0x08000000 --o STM32_test.dfu\n", exename);
    puts("Other options:\n");
    puts(" --verbose\n");

    puts("http://tomeko.net\n");
    puts("created: 2013.05.05\n");

}


/** \brief Init default options (if not exist in command line) */
void init_options(void)
{
    strcpy(fin,"");
    strcpy(fout,"out.dfu");
    starting_address = 0x08000000;
}


int main (int argc, char** argv)
{
    int c;
    long lSize;			///< filesize
    unsigned char *buffer;		///< pointer to buffer for input file
    int option_count=0; ///< command-line options counter
    FILE *fileIn;       ///< input file


    init_options();

    while (1)
    {
        static struct option long_options[] =
            {
                /* Flag-type options */
                {"verbose", no_argument,       &verbose_flag, 1
                },
                /* Non-flag options distinguished by their indices. */
                {"address",    required_argument, 0, 'a'
                },
                {"in",  required_argument, 0, 'i'
                },
                {"out",  required_argument, 0, 'o'
                },
                {0, 0, 0, 0
                }
            };
        /* getopt_long stores the option index here. */
        int option_index = 0;

        c = getopt_long (argc, argv, "abc:d:f:",
                         long_options, &option_index);

        /* Detect the end of the options. */
        if (c == -1)
        {
            if(option_count<1)
                help(argv[0]); // print help if no arguments
            break;
        }
        option_count++;

        switch (c)
        {
            case 0:
            /* If this option set a flag, do nothing else now. */
            if (long_options[option_index].flag != 0)
                break;
            printf ("option %s", long_options[option_index].name);
            if (optarg)
                printf (" with arg %s", optarg);
            printf ("\n");
            break;

            case 'a':
            //printf ("option -a with value %s\n", optarg);
            if (sscanf(optarg, "%x", &starting_address) != 1)
            {
                puts("Could not extract hex starting address from argument");
                exit(1);
            }
            break;

            case 'i':
            //printf ("option -i with value `%s'\n", optarg);
            strcpy(fin, optarg);
            break;

            case 'o':
            //printf ("option -o with value `%s'\n", optarg);
            strcpy(fout, optarg);
            break;

            case '?':
            /* getopt_long already printed an error message. */
            help(argv[0]);
            break;

            default:
            abort ();
        }
    }


    if (verbose_flag)
        puts ("verbose flag is set");

    /* Print any remaining command line arguments (not options). */
    if (optind < argc)
    {
        printf ("non-option ARGV-elements: ");
        while (optind < argc)
            printf ("%s ", argv[optind++]);
        putchar ('\n');
    }

    // start processing
    fileIn = fopen(fin, "rb"); // open input file (binary)
    if (fileIn==NULL)
    {
        puts("Error opening input file");
        exit (1);
    }

    // open output file
    fileOut = fopen(fout, "wb");
    if (fileOut==NULL)
    {
        puts("Error opening output file for writing");
        exit (1);
    }


    // obtain file size.
    fseek (fileIn , 0 , SEEK_END);
    lSize = ftell (fileIn);
    rewind (fileIn);
	printf("Creating DFU file, bin file size: %d bytes, address: 0x%08X.\n", (int)lSize, starting_address);
	if(lSize == 0 || lSize > 10*1024*1024)
	{
		puts("Error: invalid input file size.");
		exit (3);
	}

    // allocate memory to contain the whole file.
    buffer = (unsigned char*) malloc (2*sizeof(int) + lSize);
    if (buffer == NULL)
    {
    	puts("malloc for input file buffer failed (not enough memory?)");
        exit (2);
    }

    // copy the file into the buffer.
    fread (buffer + 2*sizeof(int),1,lSize,fileIn);
    fclose(fileIn);

    const char* const szDfuSe = "DfuSe";
    unsigned char ucTmp;
    unsigned short usTmp;
    unsigned int uiTmp;

    uiTmp = starting_address;
    memcpy(buffer, &uiTmp, sizeof(uiTmp));
    uiTmp = (int)lSize;
    memcpy(buffer + sizeof(int), &uiTmp, sizeof(uiTmp));

    write(szDfuSe, strlen(szDfuSe));
    ucTmp = 1;
    write(ucTmp);
    /*
    According to UM0391 this should contain whole file size.
    On the other hand files generated by STM tools have file size minus footer size here.
    Mad world...
    */
    uiTmp = 309 + lSize;
    write(uiTmp);
    uiTmp = 1;
    write(ucTmp);


    TARGETPREFIX tp;
    memset(&tp, 0, sizeof(tp));
    memcpy(tp.szSignature, TARGET_PREFIX_SIGNATURE, TARGET_PREFIX_SIGNATURE_SIZE);
    tp.dwTargetSize = 2*sizeof(int) + lSize;
    tp.dwNbElements = 1;
    write(&tp, sizeof(tp));


    write(buffer, 2*sizeof(int) + lSize);
    usTmp = 0;
    write(usTmp);
    usTmp = PID;
    write(usTmp);
    usTmp = VID;
    write(usTmp);
    usTmp = 0x011a;
    write(usTmp);
    write("UFD", strlen("UFD"));
    // The bLength field, currently fixed to 16, gives the length of the DFU Suffix itself in bytes.
    ucTmp = 16;
    write(ucTmp);
    write(crc);


	if(fileIn) fclose (fileIn);
	if(fileOut) fclose (fileOut);
	if(buffer) free (buffer);
    exit (0);
}

